#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>

#include "klib.h"

void kerr(const char *f, ...)
{
    va_list ap;
    va_start(ap, f);
    vfprintf(stderr, f, ap);
    fprintf(stderr, "\n");
    va_end(ap);
    exit(EXIT_FAILURE);
}

////

void kppol0_0(Kppol0 *w, int s)
{
    if (s < (int) sizeof (void *))
        s = sizeof (void *);
    w->n = 64 * 1024 / s;
    w->s = s;
    if (w->n < 1)
        w->n = 1;
    w->p = 0;
    #if (KPPOL0_USE != 0)
        w->a = w->u = 0;
    #endif
}

void kppol0_f(Kppol0 *w, void *p)
{
    *(void **) p = w->p;
    w->p = p;
    #if (KPPOL0_USE != 0)
        w->u--;
    #endif
}

void *kppol0_a(Kppol0 *w)
{
    char *p;
    if (w->p == 0) {
        int i;
        p = malloc(w->n * w->s);
        if (p == 0)
            kerr("Kppol0: out of memory");
        p += w->n * w->s;
        for (i = 0; i < w->n; i++) {
            p -= w->s;
            kppol0_f(w, p);
        }
        #if (KPPOL0_USE != 0)
            w->a += w->n;
            w->u += w->n;
        #endif
    }
    p = w->p;
    w->p = *(void **) p;
    #if (KPPOL0_USE != 0)
        w->u++;
    #endif
    return p;
}

////

static int kmlc0_t0[] = {
    0x4, 0x8, 0xc, 0x10,
    0x14, 0x18, 0x1c, 0x20,
    0x28, 0x30, 0x38, 0x40,
    0x50, 0x60, 0x70, 0x80,
    0xa0, 0xc0, 0xe0, 0x100,
    0x140, 0x180, 0x1c0, 0x200,
    0x280, 0x300, 0x380, 0x400,
    0x500, 0x600, 0x700, 0x800,
    0xa00, 0xc00, 0xe00, 0x1000,
    0x1400, 0x1800, 0x1c00, 0x2000,
    0x2800, 0x3000, 0x3800, 0x4000,
    0x5000, 0x6000, 0x7000, 0x8000,
    0xa000, 0xc000, 0xe000, 0x10000,
    0x14000, 0x18000, 0x1c000, 0x20000,
    0x28000, 0x30000, 0x38000, 0x40000,
    0x50000, 0x60000, 0x70000, 0x80000,
    0xa0000, 0xc0000, 0xe0000, 0x100000, // 1MB.
    0x140000, 0x180000, 0x1c0000, 0x200000,
    0x280000, 0x300000, 0x380000, 0x400000,
    0x500000, 0x600000, 0x700000, 0x800000,
    0xa00000, 0xc00000, 0xe00000, 0x1000000 // 16MB.
};

Kmlc0 kmlc0_work;

void kmlc0_0()
{
	Kmlc0 *w = &kmlc0_work;
    int i;
    w->n = sizeof (kmlc0_t0) / sizeof (kmlc0_t0[0]);
    w->t = kmlc0_t0;
    w->p = malloc(w->n * sizeof (Kppol0));
    for (i = 0; i < w->n; i++)
        kppol0_0(&w->p[i], w->t[i]);
    #if (KMLC0_CACHE != 0)
        w->cs = w->ci = 0;
    #endif
}

int kmlc0_i(int s)
{
	Kmlc0 *w = &kmlc0_work;
    int a, b, c;
    #if (KMLC0_CACHE != 0)
        if (w->cs == s) return w->ci;
    #endif
    a = b = 0;
    if (s <= w->t[0]) goto fin;
    b = w->n - 1;
    if (s > w->t[b])
        kerr("kmlc0: too large size: s=%d", s);
    #if (KMLC0_CACHE != 0)
        if (w->ci - 4 > a && s >  w->t[w->ci - 4]) a = w->ci - 4;
        if (w->ci + 4 < b && s <= w->t[w->ci + 4]) b = w->ci + 4;
    #endif
    for (;;) {
        c = (b - a) >> 1;
        if (c == 0) break;
        c += a;
        if (w->t[c] < s)
            a = c;
        else
            b = c;
    }
fin:
    #if (KMLC0_CACHE != 0)
        w->cs = s;
        w->ci = b;
    #endif
    return b;
}

void *kmlc0_a0(int i)
{
	Kmlc0 *w = &kmlc0_work;
    return kppol0_a(&w->p[i]);
}

void kmlc0_f0(void *p, int i)
{
	Kmlc0 *w = &kmlc0_work;
    kppol0_f(&w->p[i], p);
}

void *kmlc0_a1(int s)
{
    return kmlc0_a0(kmlc0_i(s));
}

void kmlc0_f1(void *p, int s)
{
	kmlc0_f0(p, kmlc0_i(s));
}

int kmlc0_s1(int s)
{
	Kmlc0 *w = &kmlc0_work;
    return w->t[kmlc0_i(s)];
}

void *kmlc0_r1(void *p, int s0, int s1)
{
	Kmlc0 *w = &kmlc0_work;
    int i0 = kmlc0_i(s0), i1 = kmlc0_i(s1);
    void *q;
    if (i0 == i1) return p;
    q = kppol0_a(&w->p[s1]);
    memcpy(q, p, s0);
    kppol0_f(&w->p[s0], p);
    return q;
}

void kmlc0_rpt(FILE *fp)
{
	Kmlc0 *w = &kmlc0_work;
    #if (KPPOL0_USE == 0)
        fprintf(fp, "kmlc0_rpt: KPPOL0_USE == 0\n");
    #else
        int i, a = 0, u = 0;
        for (i = 0; i < w->n; i++) {
            if (w->p[i].a == 0) continue;
            fprintf(fp, "kmlc0_rpt: s=0x%08x u/a=%06d/%06d\n", w->p[i].s, w->p[i].u, w->p[i].a);
            a += w->p[i].a * w->p[i].s;
            u += w->p[i].u * w->p[i].s;
        }
        fprintf(fp, "kmlc0_rpt: u/a=%d/%d\n", u, a);
    #endif
}

////

void kipol0_0(Kipol0 *w, int u, int s, int i0)
{
	w->u = u;
	w->n0 = 0;
	w->n1 = i0;
	w->s1 = s;
}

int kipol0_e(Kipol0 *w)
{
	int r = 0;
	if (w->n0 == 0 && w->n1 >= w->s1) r = -1;
	return r;
}

int kipol0_a(Kipol0 *w, void *p)
{
	int i = w->n0;
	if (i == 0) {
		if (w->n1 >= w->s1)
			kerr("kipol0: out of memory");
		return (w->n1++) | 0x80000000;
	}
	w->n0 = *(int *)(((char *) p) + i * w->u);
	return i;
}

void kipol0_f(Kipol0 *w, void *p, int i)
{
	*(int *)(((char *) p) + i * w->u) = w->n0;
	w->n0 = i;
}

////

Kps *kps;

void kps_0(Kps *w, int s, int i0)
{
	int i;
	kipol0_0(&w->ipol, sizeof (Kps_t), s, i0);	// s >= i0
	w->t = kmlc0_a1(s * sizeof (Kps_t));
	for (i = 0; i < i0; i++)
		w->t[i].s1i = 0xff;
}

int kps_a0(Kps *w)
{
	int i;
	if (kipol0_e(&w->ipol) != 0) {
		w->t = kmlc0_r1(w->t, w->ipol.s1 * sizeof (Kps_t), (w->ipol.s1 + 16) * sizeof (Kps_t));
		w->ipol.s1 += 16;
	}
	i = kipol0_a(&w->ipol, w->t);
	if ((i & 0x80000000) != 0) {
		i &= 0x7fffffff;
		w->t[i].ls = 0;
	}
	w->t[i].s1i = 0xff;
	return i;
}

void kps_f0(Kps *w, int i)
{
	w->t[i].ls++;
	kipol0_f(&w->ipol, w->t, i);
}

int kps_a1(Kps *w)
{
	int i = kps_a0(w);
	return i << 8 | w->t[i].ls;
}

int kps_a2(Kps *w, int s, int t)
{
	int i = kps_a0(w);
	if (w->t[i].s1i < 0xff)
		kerr("kps_a2: memory overload");
	w->t[i].s = s;
	w->t[i].s1i = kmlc0_i(s);
	w->t[i].p = kmlc0_a0(w->t[i].s1i);
	w->t[i].t = t;
	return i << 8 | w->t[i].ls;
}

void kps_f1(Kps *w, int is)
{
	int i = is >> 8;
	if ((is & 0xff) == w->t[i].ls)
		kps_f0(w, i);
}

void kps_f2(Kps *w, int is)
{
	Kps_t *t = w->t;
	int i = is >> 8;
	if ((is & 0xff) == t[i].ls) {
		if (t[i].s1i == 0xff)
			kerr("kps_f2: free null pointer");
		if (t[i].s1i < 0xf0)
	   		kmlc0_f0(t[i].p, t[i].s1i);
		t[i].s1i = 0xff;
		kps_f0(w, i);
	}
}

int kps_t(Kps *w, int is)
{
	int i = is >> 8, r = 0;
	if ((is & 0xff) != w->t[i].ls)
		r = -1;
	return r;
}

void *kps_p(Kps *w, int is)
{
	int i = is >> 8;
	void *r = w->t[i].p;
	if ((is & 0xff) != w->t[i].ls)
		r = 0;
	return r;
}

#if 0

long long clkusec_c = 0LL;
clock_t clkusec_t0 = 0;

long long clkusec(void)
{
    clock_t t = clock(), dt = t - clkusec_t0;
    clkusec_t0 = t;
    clkusec_c += dt * (1000000 / CLOCKS_PER_SEC);
    return clkusec_c;
}

#endif

#include <windows.h>

long long ltmusec0_c = 0LL;
int ltmusec0_t0 = -1;

long long ltmusec0(void)
{
	int t1, dt;
	SYSTEMTIME st;
	GetLocalTime(&st);
	t1 = ((st.wHour * 60 + st.wMinute) * 60 + st.wSecond) * 1000 + st.wMilliseconds;
	if (ltmusec0_t0 < 0)
		ltmusec0_t0 = t1;
	dt = t1 - ltmusec0_t0;
	if (dt < -60 * 1000)
		dt += 24 * 60 * 60 * 1000; // 60bȏ͋tsȂ낤.
	if (dt >  24 * 60 * 60 * 1000 - 60 * 1000)
		dt -= 24 * 60 * 60 * 1000; // 2359ȏ̊Ԋu͂Ȃ낤
	ltmusec0_t0 = t1;
	ltmusec0_c += dt;
    return ltmusec0_c * 1000LL;
}

long long ltmusec1_t0 = 0LL;

long long ltmusec1(void)
{
	long long t1 = ltmusec0();
	if (t1 < ltmusec1_t0)
		t1 = ltmusec1_t0;
	ltmusec1_t0 = t1;
	return t1;
}



